import { get, writable, type Writable } from "svelte/store"
import { goto } from "$app/navigation"
import { usersPagination } from "../util/itemPagination"
import {
	FullNote,
	NoteMetadata,
	User,
	VersionMetadata,
	type ApiFullNoteResponse,
	type ApiFullVersionResponse,
	type ApiNoteMetadataResponse,
	type ApiUserResponse,
	type ApiVersionMetadataResponse,
	type NewNoteResponse
} from "./model"

const API_BASE_ADDR = import.meta.env.PROD ? "/api" : "http://localhost:8080/api"
const UUID_REGEX = /^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$/i

// lifetimes of *in-memory* authentication tokens in milliseconds
const AT_EXP_MS = 15 * 60 * 1000 // 15 min.
const CSRF_EXP_MS = 12 * 60 * 60 * 1000 // 12 h.
const REFRESH_BUF = 30 * 1000 // 30 s.

// view cookie configuration (holds UNIX timestamp value of the actual refresh token cookie's expiration date)
const VIEW_COOKIE_PATH = import.meta.env.VITE_VIEW_COOKIE_PATH || "/"
const VIEW_COOKIE_DOMAIN = import.meta.env.VITE_VIEW_COOKIE_DOMAIN || "localhost"
const COOKIE_SAME_SITE = import.meta.env.VITE_COOKIE_SAME_SITE || "strict"
const COOKIE_SECURE = import.meta.env.PROD ? true : false

// some of these could just be local variables as not all of them are being used globally
export const currentUser: Writable<User | null> = writable(null)
export const currentFullNote: Writable<FullNote | null> = writable(null)
export const availableNotes: Writable<NoteMetadata[] | null> = writable(null)
export const availableUsers: Writable<User[] | null> = writable(null)
export const versionHistory: Writable<VersionMetadata[] | null> = writable(null)
export const accessToken: Writable<string | null> = writable(null)
export const csrfToken: Writable<string | null> = writable(null)
export const isPending: Writable<boolean> = writable(false)
export const cError: Writable<string | null> = writable(null)
export const cSuccess: Writable<string | null> = writable(null)

class ApiClient {
	private viewCookieName: string
	private baseUrl: string
	private lastAtUpdate = new Date(0) // refreshing the page wipes access and CSRF tokens from memory -> Rotation needed
	private lastCsrfUpdate = new Date(0)
	private refreshInProgress = false
	private activeVersion = -1
	private loadedNotesCache = new Map<string, FullNote>()
	private loadedHistoryCache = new Map<string, VersionMetadata[]>()
	private loadedVersionsCache = new Map<string, FullNote>() // key: noteID + versionID

	constructor(baseUrl: string) {
		this.baseUrl = baseUrl
		this.viewCookieName = "qnote.expires_at"
	}

	private async handleRequest<T>(
		fn: () => Promise<T>,
		options: { useBearerAuth: boolean; suspendGlobalErr: boolean }
	): Promise<T | null> {
		isPending.set(true)
		cError.set(null)
		cSuccess.set(null)

		// NOTE: if `handleResponse` is used, errors thrown from it will be caught here

		try {
			if (options.useBearerAuth) {
				try {
					await this.checkAndRefreshAccessToken()
				} catch (err) {
					console.log("[REQ] Refresh attempt not successful")
					await this.handleLocalLogout()
					throw new Error("Session expired, please authenticate again.")
				}
			}

			return await fn()
		} catch (err) {
			const errMsg = err instanceof Error ? err.message : "Unknown error"

			// the suspension option is handy when we want to display the error inside a modal instead of in a global notification
			if (options.suspendGlobalErr) {
				// throw the same error to the next handler (should be handled inside the caller component)
				throw new Error(errMsg)
			} else {
				cError.set(errMsg)
			}

			console.log(`[ERR] ${errMsg}`)
		} finally {
			isPending.set(false)
		}

		return null
	}

	// should be attached to routes that handle authentication with the bearer token (access token)
	private async handleResponse<T>(
		response: Response,
		options: { useBearerAuth: boolean }
	): Promise<T> {
		if (!response.ok) {
			if (response.status === 401 && options.useBearerAuth) {
				// this should never happen due to the token expiration checks we make client-side,
				// but it's still good to have as a fallback
				try {
					console.log("[RES] Unexpected 401 caught, attempting to refresh...")
					await this.checkAndRefreshAccessToken()
				} catch (err) {
					const errMsg = err instanceof Error ? err.message : "Unknown error"
					await this.handleLocalLogout()
					throw new Error(errMsg)
				}
			}

			// capitalize the error message and display (only if not 401)
			const { error } = await response.json()
			const dError = error[0].toUpperCase() + error.substr(1).toLowerCase() + "."

			throw new Error(dError)
		}

		return await response.json()
	}

	private async checkAndRefreshAccessToken(): Promise<void> {
		// notably we must check whether we even have an authentication cookie present
		// as that's obviously required for completing the token rotation procedure
		if (this.refreshInProgress) {
			return
		} else if (!this.isAuthCookieValid()) {
			throw new Error("Session expired, please authenticate again.")
		}

		const timeSinceUpdate = Date.now() - this.lastAtUpdate.getTime()
		const needsRefresh = timeSinceUpdate > AT_EXP_MS - REFRESH_BUF

		if (needsRefresh) {
			console.log(`[AUTH] Running token refresh attempt (timeSinceUpdate=${timeSinceUpdate})`)

			this.refreshInProgress = true
			await this.refreshAccessToken()
			this.refreshInProgress = false
		}
	}

	private async refreshAccessToken(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/cookie/refresh`, {
					method: "POST",
					headers: {
						...(await this.getCsrfHeader())
					},
					credentials: "include"
				})

				if (!response.ok) {
					await this.handleLocalLogout()
					throw new Error("Refreshing token failed.")
				}

				const { access_token: newToken } = await response.json()
				accessToken.set(newToken)
				this.lastAtUpdate = new Date()
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	private async getCsrfHeader(): Promise<HeadersInit> {
		let token = get(csrfToken)
		const timeSinceUpdate = Date.now() - this.lastCsrfUpdate.getTime()
		const needsRefresh = timeSinceUpdate > CSRF_EXP_MS - REFRESH_BUF

		if (!token || needsRefresh) {
			console.log("[AUTH] Refreshing CSRF token")
			await this.refreshCsrfToken()
			token = get(csrfToken)
		}

		return { "X-Csrf-Token": token || "" }
	}

	private async refreshCsrfToken(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/cookie/csrf`, {
					credentials: "include"
				})

				if (!response.ok) {
					await this.handleLocalLogout()
					throw new Error("Fetching CSRF token failed.")
				}

				const newToken = response.headers.get("X-Csrf-Token")
				csrfToken.set(newToken)
				this.lastCsrfUpdate = new Date()
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	private async handleLocalLogout(): Promise<void | null> {
		this.lastAtUpdate = new Date(0)
		this.lastCsrfUpdate = new Date(0)

		currentUser.set(null)
		accessToken.set(null)
		csrfToken.set(null)

		this.deleteViewCookie()

		await goto("/login")
	}

	private getAuthHeader(): HeadersInit {
		const token = get(accessToken)
		return { Authorization: `Bearer ${token}` }
	}

	private isAuthCookieValid(): boolean {
		const now = new Date()
		const viewCookie = this.getCookieValue(this.viewCookieName)

		if (!viewCookie) {
			console.log("[AUTH] View cookie not found")
			return false
		}

		const value = viewCookie.split("=")[1]
		const expirationTimestamp = parseInt(value, 10)

		if (isNaN(expirationTimestamp)) {
			console.log(`[ERR] Invalid view cookie expiration timestamp: ${value}`)
			return false
		}

		return now < new Date(expirationTimestamp * 1000)
	}

	private deleteViewCookie() {
		const viewCookie = this.getCookieValue(this.viewCookieName)
		if (!viewCookie) {
			return
		}

		// overwrite the view cookie with details that match with the real cookie
		document.cookie = `${this.viewCookieName}=;path=${VIEW_COOKIE_PATH};domain=${VIEW_COOKIE_DOMAIN};expires=Thu, 01 Jan 1970 00:00:00 GMT;${COOKIE_SECURE ? "secure;" : ""}sameSite=${COOKIE_SAME_SITE}`
	}

	private getCookieValue(cookieName: string): string | null {
		const cookieString = document.cookie
		const cookies = cookieString.split(";").map((cookie) => cookie.trim())
		const cookieValue = cookies.find((cookie) => cookie.startsWith(`${cookieName}=`))

		if (!cookieValue) {
			return null
		}

		return cookieValue
	}

	private joinDeserializedVersion(
		noteID: string,
		apiResponse: ApiFullVersionResponse
	): FullNote | null {
		// cache lookups are safe here due to this always being called *after* fetching the actual `FullNote`
		const cachedNote = this.loadedNotesCache.get(noteID)

		if (!cachedNote) {
			return null
		}

		return cachedNote.joinWithVersionResponse(apiResponse)
	}

	public async register(username: string, password: string): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/signup`, {
					method: "POST",
					headers: { "Content-Type": "application/json" },
					body: JSON.stringify({ username, password })
				})

				// can't overwrite the function parameter
				const data = await this.handleResponse<{
					id: string
					username: string
				}>(response, { useBearerAuth: false })

				console.log(`[USER] Registration of user '${data.username}' successful`)
				cSuccess.set(`User created successfully.`)

				await goto("/login")
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	public async login(username: string, password: string): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/login`, {
					method: "POST",
					headers: { "Content-Type": "application/json" },
					body: JSON.stringify({ username, password }),
					credentials: "include"
				})

				const { access_token: token } = await this.handleResponse<{
					access_token: string
				}>(response, { useBearerAuth: false })

				accessToken.set(token)
				this.lastAtUpdate = new Date()

				goto("/")
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	public async logout(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/logout`, {
					method: "POST",
					headers: {
						...this.getAuthHeader()
					},
					credentials: "include"
				})

				if (response.status === 204) {
					console.log("[USER] Logout successful")
					await this.handleLocalLogout()
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
				await this.handleLocalLogout()
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getUserCheckAuth(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/me`, {
					headers: {
						...this.getAuthHeader()
					}
				})
				const data = await this.handleResponse<ApiUserResponse>(response, { useBearerAuth: true })
				const user = new User(data)

				currentUser.set(user)
			},
			{ useBearerAuth: true, suspendGlobalErr: true }
		)
	}

	public async updateCurrentUserPassword(
		oldPassword: string,
		newPassword: string
	): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const data = {
					old_password: oldPassword,
					new_password: newPassword
				}
				const response = await fetch(`${this.baseUrl}/auth/owner`, {
					method: "PUT",
					headers: {
						...this.getAuthHeader(),
						"Content-Type": "application/json"
					},
					credentials: "include",
					body: JSON.stringify(data)
				})

				const { access_token: token, user } = await this.handleResponse<{
					access_token: string
					user: User
				}>(response, { useBearerAuth: true })

				accessToken.set(token)
				currentUser.set(user || null)
				this.lastAtUpdate = new Date()
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // error displayed inside the settings modal
		)
	}

	public async deleteCurrentUser(password: string): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/owner`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader(),
						"Content-Type": "application/json"
					},
					body: JSON.stringify({ password })
				})

				if (response.status === 204) {
					console.log("[USER] Deletion successful")
					await this.handleLocalLogout()
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
				await this.handleLocalLogout()
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // error displayed inside the settings modal
		)
	}

	public async adminListAll(): Promise<User[] | null> {
		const user = get(currentUser)
		if (!user || !user.isAdmin) {
			throw new Error("Admin privileges required.")
		}

		return this.handleRequest(
			async () => {
				const params = new URLSearchParams()
				params.append("limit", `${get(usersPagination).pageSize}`)
				params.append("offset", `${get(usersPagination).currentPage}`)
				const response = await fetch(`${this.baseUrl}/auth/admin/all?${params}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				let users: User[] = []
				const data = await this.handleResponse<ApiUserResponse[]>(response, { useBearerAuth: true })

				if (users) {
					users = User.fromApiResponseArray(data)
				}

				console.log(`[ADMIN] Got ${users.length} user results`)

				return users
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // error displayed inside the settings modal
		)
	}

	public async adminDeleteUser(userID: string): Promise<void | null> {
		const user = get(currentUser)
		if (!user || !user.isAdmin) {
			throw new Error("Admin privileges required.")
		}

		if (!UUID_REGEX.test(userID)) {
			throw new Error("Invalid user ID format.")
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/admin/${userID}`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader()
					}
				})

				if (response.status === 204) {
					console.log(`[ADMIN] Deletion of user '${userID}' successful`)
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // error displayed inside the settings modal
		)
	}

	public async listNotes(): Promise<NoteMetadata[] | null> {
		return this.handleRequest(
			async () => {
				const params = new URLSearchParams()
				params.append("limit", `${get(usersPagination).pageSize}`)
				params.append("offset", `${get(usersPagination).currentPage}`)
				const response = await fetch(`${this.baseUrl}/notes?${params}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				let notes: NoteMetadata[] = []
				const data = await this.handleResponse<ApiNoteMetadataResponse[]>(response, {
					useBearerAuth: true
				})

				if (data) {
					notes = NoteMetadata.fromApiResponseArray(data)
				}

				console.log(`[NOTE] Got ${notes.length} note metadata results`)

				return notes
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async createNote(): Promise<NewNoteResponse | null> {
		// NOTE: the initial note version doesn't allow any user input, the first user-made modification
		//       is applied through the version creation endpoint

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes`, {
					method: "POST",
					headers: {
						...this.getAuthHeader()
					}
				})
				return await this.handleResponse<NewNoteResponse>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getActiveFullNote(noteID: string, fetchRemote: boolean): Promise<FullNote | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		// attempt cache lookup only if we didn't just push new updates
		if (!fetchRemote) {
			const cachedNote = this.loadedNotesCache.get(noteID)
			if (cachedNote != null) {
				this.activeVersion = cachedNote.versionNumber
				return cachedNote
			}
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const data = await this.handleResponse<ApiFullNoteResponse>(response, {
					useBearerAuth: true
				})
				const note = new FullNote(data)

				console.log(`[CACHE] Storing ${noteID}`)
				this.loadedNotesCache.set(noteID, note)
				this.activeVersion = note.versionNumber

				return note
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async deleteNote(noteID: string): Promise<void | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader()
					}
				})

				if (response.status === 204) {
					console.log("[NOTE] Deletion successful")
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getNoteHistory(
		noteID: string,
		fetchRemote: boolean
	): Promise<VersionMetadata[] | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		if (!fetchRemote) {
			const cachedVersions = this.loadedHistoryCache.get(noteID)
			if (cachedVersions != null) {
				return cachedVersions
			}
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}/versions`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const data = await this.handleResponse<ApiVersionMetadataResponse[]>(response, {
					useBearerAuth: true
				})
				const versions = VersionMetadata.fromApiResponseArray(data, this.activeVersion)

				this.loadedHistoryCache.set(noteID, versions)
				console.log(`[VER] Got and cached ${versions.length} version metadata results`)

				return versions
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async createVersion(noteID: string, title: string, content: string): Promise<void | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		// NOTE: title's length limit is applied in the UI component, so we don't need to worry about it here

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}/versions`, {
					method: "POST",
					headers: {
						...this.getAuthHeader(),
						"Content-Type": "application/json"
					},
					body: JSON.stringify({ title, content })
				})

				if (response.status === 204) {
					console.log("[VER] Creation successful")
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getFullVersion(
		noteID: string,
		versionID: string,
		isActiveVersion: boolean
	): Promise<FullNote | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		if (!UUID_REGEX.test(versionID)) {
			throw new Error("Invalid version ID format.")
		}

		if (isActiveVersion) {
			// the active version will always get cached when the note is initially selected and loaded
			const cachedActiveNote = this.loadedNotesCache.get(noteID)
			if (cachedActiveNote !== undefined) {
				return cachedActiveNote
			}
		}

		// NOTE: no need to explicitly prevent attempting a cache hit as versions aren't editable

		const cachedVersion = this.loadedVersionsCache.get(noteID + versionID)
		if (cachedVersion != null) {
			return cachedVersion
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}/${versionID}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const data = await this.handleResponse<ApiFullVersionResponse>(response, {
					useBearerAuth: true
				})
				const version = this.joinDeserializedVersion(noteID, data)

				if (!version) {
					return null
				}

				console.log(`[CACHE] Storing [${noteID}, ${versionID}]`)
				this.loadedVersionsCache.set(noteID + versionID, version)

				return version
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}
}

export const apiClient = new ApiClient(API_BASE_ADDR)
